# coding=utf-8

from __future__ import division
from pymycobot.common import ProtocolCode
from pymycobot.myarm_api import MyArmAPI


class MyArmC(MyArmAPI):

    def __init__(self, port, baudrate=1000000, timeout=0.1, debug=False):
        super(MyArmC, self).__init__(port, baudrate, timeout, debug)

    def is_tool_btn_clicked(self, mode=1):
        """get the end button status
        Args:
            1: atom
            2: gripper red button
            3: gripper blue button
            254: get all button status
        Returns:
            list[int]: 0/1, 1: press, 0: no press
        """
        if not isinstance(mode, int):
            raise TypeError('mode must be int')

        if mode not in [1, 2, 3, 254]:
            raise ValueError('mode must be 1, 2, 3 or 254')

        return self._mesg(ProtocolCode.GET_ATOM_PRESS_STATUS, mode, has_reply=True)




import logging
import threading
import time
import struct
from serial import Serial


class CommandGenre(object):
    SET_SERVO_DIRECTION = 0XA0
    GET_SERVO_DIRECTION = 0XA2

    SET_SERVO_SPEED = 0XA5
    GET_SERVO_SPEED = 0XA3

    # WRITE_SERVO_ANGLE = 0XA6
    # WRITE_SERVO_STEP = 0XA7

    READ_FIRMWARE_VERSION = 0XAA


class Command(object):
    HEADER = 0XFF

    def __init__(self, genre, address, check_digit, params, length=None):
        self.__header = [0xff, 0xff]
        self.__length = length or len(params)
        self.__address = address
        self.__params = params
        self.__genre = genre
        self.__check_digit = check_digit

    @property
    def genre(self):
        return self.__genre

    def to_bytes(self):
        return bytes([*self.__header, self.__address, self.__length, *self.__params, self.__genre, *self.__check_digit])

    def get_params(self):
        return self.__params[0] if self.__length == 1 else self.__params

    def __str__(self):
        return " ".join(map(lambda bit: hex(bit).upper(), self.to_bytes()))

    def __bytes__(self):
        return self.to_bytes()

    @classmethod
    def packing(cls, genre: int, addr: int, *params):
        check_code = cls.check_digit(genre, params)
        return cls(genre=genre, address=addr, check_digit=(check_code, ), params=(*params,))

    @classmethod
    def parsing(cls, buffer: bytes):
        # header, header, addr, length, params, genre, check_bits
        if len(buffer) < 6:
            return None
        if buffer[0] != cls.HEADER or buffer[1] != cls.HEADER:
            return None
        length = buffer[3]
        return cls(genre=buffer[-2], address=buffer[2], length=length, params=(*buffer[4:4+length], ), check_digit=buffer[4+length:4+length+1])

    @classmethod
    def unpack_args(cls, *parameters):
        bits_pack_list = []
        for param in parameters:
            pair = struct.pack('>h', param)
            if len(pair) == 2:
                bits_pack_list.extend(list(pair))
            else:
                bits_pack_list.clear()
        print(bits_pack_list)
        return bits_pack_list

    @classmethod
    def check_digit(cls, genre, params):
        """
        Calculate the check-code for the command.
        :param genre: int, function genre
        :param params: bytes, function parameters
        :return: int, check-code
        """
        return sum([genre, *params]) & 0xff

    @classmethod
    def has_header(cls, buffer: bytes):
        """
        Check if the buffer contains a header.
        :param buffer:
        :return:
        """
        if len(buffer) < 2:
            return False
        return buffer[0] == cls.HEADER and buffer[1] == cls.HEADER


class SerialProtocol(object):

    def __init__(self, comport, baudrate, timeout=0.5):
        self._comport = comport
        self._baudrate = baudrate
        self._timeout = timeout
        self._serial_port = Serial(port=comport, baudrate=baudrate, timeout=timeout)
        self._serial_port.rts = True
        self._serial_port.dtr = True

    def open(self):
        if self._serial_port.is_open is False:
            self._serial_port.open()

    def close(self):
        if self._serial_port.is_open is True:
            self._serial_port.close()

    def write(self, data: bytes):
        self._serial_port.write(data)
        self._serial_port.flush()

    def read(self, size=None):
        return self._serial_port.read(size or self._serial_port.in_waiting)

    def flush(self):
        self._serial_port.flush()


class ConveyorAPI(SerialProtocol):
    class MotorModel:
        STEPPER_MOTOR_42 = 0x30
        STEPPER_MOTOR_57 = 0x31

    def __init__(self, comport, baudrate="115200", timeout=0.1, debug=False):
        super().__init__(comport, baudrate, timeout)
        self._debug = debug
        self.open()
        self._mutex = threading.Lock()
        self._log = logging.getLogger("conveyor_api")
        self._log.setLevel(logging.DEBUG if debug else logging.INFO)
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG if debug else logging.INFO)
        formatter = logging.Formatter(
            fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )
        handler.setFormatter(formatter)
        self._log.addHandler(handler)

    def _wait_for_reply(self, timeout=None):
        buffers = b""
        timeout = timeout or self._timeout
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < timeout:
            if self._serial_port.in_waiting <= 0:
                continue
            buffers = self.read(self._serial_port.in_waiting)
            if not Command.has_header(buffers):
                continue
            break

        return Command.parsing(buffers) if Command.has_header(buffers) else None

    def _merge(self, genre, address, *parameters, has_reply=False):
        command = Command.packing(genre, address, *parameters)
        self._log.debug(f"write > {command}")
        with self._mutex:
            self.write(command.to_bytes())

            if not has_reply:
                return

            reply_command = self._wait_for_reply(timeout=0.07)  # WaitForAReply
            self._log.debug(f"read  < {reply_command}")
            if not reply_command:
                return None

            result = reply_command.get_params()
            if reply_command.genre == CommandGenre.READ_FIRMWARE_VERSION:
                return result / 10
            return result

    def set_motor_direction(self, direction, motor_model=MotorModel.STEPPER_MOTOR_57):
        """Modify the direction of movement of the conveyor belt"""
        if direction not in (0, 1):
            raise ValueError("direction must be 0 or 1")
        self._merge(CommandGenre.SET_SERVO_DIRECTION, motor_model, direction)

    def get_motor_direction(self, motor_model=MotorModel.STEPPER_MOTOR_57):
        """Get the direction of movement of the conveyor belt"""
        return self._merge(CommandGenre.GET_SERVO_DIRECTION, motor_model, has_reply=True)

    def get_motor_speed(self, motor_model=MotorModel.STEPPER_MOTOR_57):
        """Get the speed of the conveyor belt"""
        return self._merge(CommandGenre.GET_SERVO_SPEED, motor_model, has_reply=True)

    def set_motor_speed(self, status, speed: int, motor_model=MotorModel.STEPPER_MOTOR_57):
        """Modify the speed of the conveyor belt"""
        if status not in (0, 1):
            raise ValueError("status must be 0 or 1")

        if not 0 <= speed <= 100:
            raise ValueError("speed must be in range [0, 100]")
        return self._merge(CommandGenre.SET_SERVO_SPEED, motor_model, status, speed)

    # def write_motor_angle(self, carry, angle: int, speed,  motor_model=MotorModel.STEPPER_MOTOR_57):
    #     return self._merge(CommandGenre.WRITE_SERVO_ANGLE, motor_model, carry, angle, speed)
    #
    # def write_motor_step(self, carry, step: int, speed, direction, motor_model=MotorModel.STEPPER_MOTOR_57):
    #     return self._merge(CommandGenre.WRITE_SERVO_STEP, motor_model, carry, step, speed, direction)

    def read_firmware_version(self, motor_model=MotorModel.STEPPER_MOTOR_57):
        """Get the firmware version of the conveyor belt"""
        return self._merge(CommandGenre.READ_FIRMWARE_VERSION, motor_model, has_reply=True)




# coding=utf-8

import numpy as np

def CvtRotationMatrixToEulerAngle(pdtRotationMatrix):
    pdtEulerAngle = np.zeros(3)

    pdtEulerAngle[2] = np.arctan2(pdtRotationMatrix[1, 0], pdtRotationMatrix[0, 0])

    fCosRoll = np.cos(pdtEulerAngle[2])
    fSinRoll = np.sin(pdtEulerAngle[2])

    pdtEulerAngle[1] = np.arctan2(-pdtRotationMatrix[2, 0], (fCosRoll * pdtRotationMatrix[0, 0]) + (fSinRoll * pdtRotationMatrix[1, 0]))
    pdtEulerAngle[0] = np.arctan2((fSinRoll * pdtRotationMatrix[0, 2]) - (fCosRoll * pdtRotationMatrix[1, 2]), (-fSinRoll * pdtRotationMatrix[0, 1]) + (fCosRoll * pdtRotationMatrix[1, 1]))

    return pdtEulerAngle

def CvtEulerAngleToRotationMatrix(ptrEulerAngle):
    ptrSinAngle = np.sin(ptrEulerAngle)
    ptrCosAngle = np.cos(ptrEulerAngle)

    ptrRotationMatrix = np.zeros((3, 3))
    ptrRotationMatrix[0, 0] = ptrCosAngle[2] * ptrCosAngle[1]
    ptrRotationMatrix[0, 1] = ptrCosAngle[2] * ptrSinAngle[1] * ptrSinAngle[0] - ptrSinAngle[2] * ptrCosAngle[0]
    ptrRotationMatrix[0, 2] = ptrCosAngle[2] * ptrSinAngle[1] * ptrCosAngle[0] + ptrSinAngle[2] * ptrSinAngle[0]
    ptrRotationMatrix[1, 0] = ptrSinAngle[2] * ptrCosAngle[1]
    ptrRotationMatrix[1, 1] = ptrSinAngle[2] * ptrSinAngle[1] * ptrSinAngle[0] + ptrCosAngle[2] * ptrCosAngle[0]
    ptrRotationMatrix[1, 2] = ptrSinAngle[2] * ptrSinAngle[1] * ptrCosAngle[0] - ptrCosAngle[2] * ptrSinAngle[0]
    ptrRotationMatrix[2, 0] = -ptrSinAngle[1]
    ptrRotationMatrix[2, 1] = ptrCosAngle[1] * ptrSinAngle[0]
    ptrRotationMatrix[2, 2] = ptrCosAngle[1] * ptrCosAngle[0]

    return ptrRotationMatrix

def transformation_matrix_from_parameters(rotation_matrix, translation_vector):
    """
    Create a 4x4 homogeneous transformation matrix.

    :param rotation_matrix: 3x3 numpy array representing rotation.
    :param translation_vector: 1x3 numpy array representing translation.
    :return: 4x4 numpy array representing the transformation matrix.
    """
    T = np.eye(4)
    T[:3, :3] = rotation_matrix
    T[:3, 3] = translation_vector
    return T

def get_flange_pose(flange_matrix, tool_pose):
    """
    Convert the pose from flange coordinate system to tool coordinate system.

    :param flange_to_tool_matrix: 4x4 numpy array, transformation matrix from flange to tool.
    :param flange_pose: 4x4 numpy array, pose in flange coordinate system.
    :return: 4x4 numpy array, pose in tool coordinate system.
    """
    return np.dot(flange_matrix, np.linalg.inv(tool_pose))

def get_tool_pose(flange_matrix, tool_pose):
    """
    Convert the pose from tool coordinate system to flange coordinate system.

    :param flange_to_tool_matrix: 4x4 numpy array, transformation matrix from flange to tool.
    :param tool_pose: 4x4 numpy array, pose in tool coordinate system.
    :return: 4x4 numpy array, pose in flange coordinate system.
    """
    return np.dot(flange_matrix, tool_pose)


def flangeToTool(current_coords, tool_matrix):
    # Example pose in flange coordinate system
    flange_pose = np.eye(4)
    flange_pose[:3, :3] = CvtEulerAngleToRotationMatrix(current_coords[3:6] * np.pi/180.0)
    flange_pose[:3, 3] = current_coords[:3]  # Example translation

    # Switch to tool coordinate system
    tool_pose = get_tool_pose(flange_pose, tool_matrix)
    tool_coords = np.concatenate((tool_pose[:3,3].T, CvtRotationMatrixToEulerAngle(tool_pose[:3,:3]) * 180/np.pi))

    return tool_coords

def toolToflange(tool_coords, tool_matrix):

    # Example pose in tool coordinate system
    tool_pose = np.eye(4)
    tool_pose[:3, :3] = CvtEulerAngleToRotationMatrix(tool_coords[3:6] * np.pi/180.0)
    tool_pose[:3, 3] = tool_coords[:3] # Example translation

    # Switch to tool coordinate system
    flange_pose = get_flange_pose(tool_pose, tool_matrix)
    
    flange_coords = np.concatenate((flange_pose[:3,3].T, CvtRotationMatrixToEulerAngle(flange_pose[:3,:3]) * 180/np.pi ))
    return flange_coords


#!/usr/bin/python

from socket import socket, AF_INET, SOCK_STREAM
import sys
from enum import Enum
import base64
import hashlib
import math
import numpy as np
from multiprocessing import Lock
import logging
from pymycobot.log import setup_logging
from pymycobot.tool_coords import *


COORDS_EPSILON = 0.50


class Axis(Enum):
    X = 0
    Y = 1
    Z = 2
    RX = 3
    RY = 4
    RZ = 5


class Joint(Enum):
    J1 = 0
    J2 = 1
    J3 = 2
    J4 = 3
    J5 = 4
    J6 = 5


class JogMode(Enum):
    JOG_JOINT = 0
    JOG_TELEOP = 1


mutex = Lock()


class ElephantRobot(object):
    def __init__(self, host, port, debug=False):
        # setup connection
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)
        self.BUFFSIZE = 8 * 1024 * 1024
        self.ADDR = (host, port)
        self.tcp_client = socket(AF_INET, SOCK_STREAM)
        self.is_client_started = False
        tool_reference = np.array([0, 0, 0, 0, 0, 0])
        self.tool_matrix = self.set_tool_reference(tool_reference)

    def get_ip(self):
        """Returns tuple of (IP address or hostname, port) of robot with socket server.

        Returns:
            str: IP address or hostname
        """
        return self.ADDR

    def set_ip(self, host, port):
        """Sets IP address or hostname and port of robot with socket server.
        Cannot set or change IP address or hostname or port of already started client.
        Need to stop client before setting IP address/hostname/port.

        Args:
            host (str): IP address or hostname
            port (int): port

        Returns:
            bool: True if success, False otherwise
        """
        if self.is_client_started:
            print(
                "Error: Cannot change IP if client is started. Stop client to change IP/Host/Port"
            )
            return False
        self.ADDR = (host, port)
        return True

    def start_client(self):
        try:
            self.tcp_client.connect(self.ADDR)
            self.is_client_started = True
            return True
        except Exception as e:
            print(e)
            return False

    def stop_client(self):
        self.tcp_client.close()
        self.is_client_started = False

    def send_command(self, command):
        with mutex:
            self.tcp_client.send(command.encode())
            recv_data = self.tcp_client.recv(self.BUFFSIZE).decode()
            res_str = str(recv_data)
            if self.debug:
                print("recv = " + res_str)
            res_arr = res_str.split(":")
            if len(res_arr) == 2:
                return res_arr[1]
            else:
                return ""

    def string_to_coords(self, data):
        data = data.replace("[", "")
        data = data.replace("]", "")
        data_arr = data.split(",")
        if len(data_arr) == 6:
            try:
                coords_1 = round(float(data_arr[0]), 3)
                coords_2 = round(float(data_arr[1]), 3)
                coords_3 = round(float(data_arr[2]), 3)
                coords_4 = round(float(data_arr[3]), 3)
                coords_5 = round(float(data_arr[4]), 3)
                coords_6 = round(float(data_arr[5]), 3)
                coords = [coords_1, coords_2, coords_3, coords_4, coords_5, coords_6]
                return coords
            except:
                return self.invalid_coords()
        return self.invalid_coords()

    def string_to_double(self, data):
        try:
            val = round(float(data), 3)
            return val
        except:
            return -9999.99

    def float_equal(self, a, b, epsilon=COORDS_EPSILON):
        """Compares 2 floats with given epsilon precision.

        Args:
            a (float): 1st float to compare
            b (float): 2nd float to compare
            epsilon (float): precision (epsilon) to compare

        Returns:
            bool: True if floats are considered equal with given precision,
                  False otherwise
        """
        return math.fabs(a - b) < epsilon

    def coords_equal(self, coords_1, coords_2):
        """Checks if specified coords are equal.

        Args:
            coords_1 (list): first coords to compare
            coords_2 (list): second coords to compare

        Returns:
            bool: True if coords are equal, False otherwise.
        """
        return (
            self.float_equal(coords_1[Axis.X.value], coords_2[Axis.X.value])
            and self.float_equal(coords_1[Axis.Y.value], coords_2[Axis.Y.value])
            and self.float_equal(coords_1[Axis.Z.value], coords_2[Axis.Z.value])
            and self.float_equal(coords_1[Axis.RX.value], coords_2[Axis.RX.value])
            and self.float_equal(coords_1[Axis.RY.value], coords_2[Axis.RY.value])
            and self.float_equal(coords_1[Axis.RZ.value], coords_2[Axis.RZ.value])
        )

    def angles_equal(self, angles_1, angles_2):
        """Checks if specified angles are equal.

        Args:
            angles_1 (list): first angles to compare
            angles_2 (list): second angles to compare

        Returns:
            bool: True if angles are equal, False otherwise.
        """
        return (
            self.float_equal(angles_1[Joint.J1.value], angles_2[Joint.J1.value])
            and self.float_equal(angles_1[Joint.J2.value], angles_2[Joint.J2.value])
            and self.float_equal(angles_1[Joint.J3.value], angles_2[Joint.J3.value])
            and self.float_equal(angles_1[Joint.J4.value], angles_2[Joint.J4.value])
            and self.float_equal(angles_1[Joint.J5.value], angles_2[Joint.J5.value])
            and self.float_equal(angles_1[Joint.J6.value], angles_2[Joint.J6.value])
        )

    def string_to_int(self, data):
        try:
            val = int(data)
            return val
        except:
            return -9999

    def invalid_coords(self):
        coords = [-1.0, -2.0, -3.0, -4.0, -1.0, -1.0]
        return coords

    def detect_robot(self):
        return int(self.get_analog_in(55))

    def start_robot(self):
        """Starts robot.

        Returns:
            bool: currently always returns True
        """
        command = "start_robot()\n"
        res = self.send_command(command)
        return True

    def recover_robot(self):
        """Recovers robot after collision.

        Returns:
            bool: currently always returns True
        """
        return self.start_robot()

    def get_angles(self):
        command = "get_angles()\n"
        res = self.send_command(command)
        return self.string_to_coords(res)

    def get_coords(self):
        command = "get_coords()\n"
        res = self.send_command(command)
        return self.string_to_coords(res)

    def get_speed(self):
        command = "get_speed()\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def _power_on(self):
        command = "power_on()\n"
        res = self.send_command(command)
        return True

    def _power_off(self):
        command = "power_off()\n"
        res = self.send_command(command)
        return True

    def _state_on(self):
        command = "state_on()\n"
        self.send_command(command)

    def _state_off(self):
        command = "state_off()\n"
        self.send_command(command)

    def state_check(self):
        command = "state_check()\n"
        res = self.send_command(command)
        return res == "1"

    def check_running(self):
        command = "check_running()\n"
        res = self.send_command(command)
        return res == "1"

    def is_in_position(self, coords, jog_mode):
        """Returns True if current position equals passed coords.

        Args:
            coords (list[float]): coords or angles
            jog_mode (JogMode): JogMode enum value

        Returns:
            bool: True if robot is in passed coords, False otherwise
        """
        if type(jog_mode) is not JogMode:
            raise TypeError("jog_mode should be of type JogMode")
        if jog_mode == JogMode.JOG_TELEOP:
            return self.coords_equal(self.get_coords(), coords)
        else:
            return self.angles_equal(self.get_angles(), coords)

    def set_free_move_mode(self, on=True):
        """Enables or disables free move mode.

        Args:
            on (bool, optional): True to enable free move mode,
                                 False to disable free move mode.
                                 Defaults to True.
        """
        command = "set_free_move(" + str(int(on)) + ")\n"
        self.send_command(command)

    def set_payload(self, payload):
        """Sets payload value on the robot.

        Args:
            payload (float): payload value (0.0~2.0)
        """
        command = "set_payload(" + str(payload) + ")\n"
        self.send_command(command)

    def get_payload(self):
        """Returns current payload value.

        Returns:
            float: payload value
        """
        return self.get_analog_out(53)

    def upload_file(self, local_filename, remote_filename):
        """Upload local file to robot via socket. Permissions are checked before
           upload so it cannot overwrite system files (owned by root).

        Args:
            local_filename (str): absolute or relative path to a file to send
            remote_filename (str): absolute or relative file path to where
                                   to upload a file. If path is relative, it is
                                   relative to ~/RoboFlow/upload folder.

        Returns:
            str: ok if success or error message otherwise
        """
        with open(local_filename, "rb") as f:
            content = f.read()
        content_base64 = base64.b64encode(content).decode()
        content_sha256 = hashlib.sha256(content).hexdigest()
        command = "upload_file({},{},{})".format(
            content_base64, remote_filename, content_sha256
        )
        res = self.send_command(command)
        return res

    def read_text_file(self, remote_filename):
        command = f"read_text_file({remote_filename})"
        res = self.send_command(command)
        return res

    def program_open(self, file_path):
        command = "program_open(" + file_path + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def program_run(self, start_line):
        command = "program_run(" + str(start_line) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def read_next_error(self):
        command = "read_next_error()\n"
        res = self.send_command(command)
        return res

    def clear_all_errors(self):
        while self.read_next_error() != "":
            pass

    def write_coords(self, coords, speed):
        command = "set_coords("
        for item in coords:
            command += str(round(item, 3)) + ","
        command += str(speed) + ")\n"
        self.send_command(command)

    def write_coord(self, axis, value, speed):
        coords = self.get_coords()
        if coords != self.invalid_coords():
            coords[axis] = value
            self.write_coords(coords, speed)

    def write_angles(self, angles, speed):
        command = "set_angles("
        for item in angles:
            command += str(round(item, 3)) + ","
        command += str(speed) + ")\n"
        self.send_command(command)

    def write_angle(self, joint, value, speed):
        angles = self.get_angles()
        if angles != self.invalid_coords():
            angles[joint] = value
            self.write_angles(angles, speed)

    def set_speed(self, percentage):
        command = "set_speed(" + str(percentage) + ")\n"
        self.send_command(command)

    def set_carte_torque_limit(self, axis_str, value):
        command = "set_torque_limit(" + axis_str + "," + str(value) + ")\n"
        self.send_command(command)

    def set_upside_down(self, up_down):
        up = "1"
        if up_down:
            up = "0"
        command = "set_upside_down(" + up + ")\n"
        self.send_command(command)

    def task_stop(self):
        command = "task_stop()\n"
        self.send_command(command)

    def jog_angle(self, joint_str, direction, speed):
        command = (
            "jog_angle(" + joint_str + "," + str(direction) + "," + str(speed) + ")\n"
        )
        self.send_command(command)

    def jog_coord(self, axis_str, direction, speed):
        command = (
            "jog_coord(" + axis_str + "," + str(direction) + "," + str(speed) + ")\n"
        )
        self.send_command(command)

    def get_digital_in(self, pin_number):
        command = "get_digital_in(" + str(pin_number) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_digital_out(self, pin_number):
        command = "get_digital_out(" + str(pin_number) + ")\n"
        print(command)
        res = self.send_command(command)
        return self.string_to_int(res)

    def set_digital_out(self, pin_number, pin_signal):
        command = "set_digital_out(" + str(pin_number) + "," + str(pin_signal) + ")\n"
        self.send_command(command)

    def get_analog_in(self, pin_number):
        """Returns specified analog input pin value.

        Args:
            pin_number (int): pin number (0-63)

        Returns:
            float: pin value
        """
        command = "get_analog_in(" + str(pin_number) + ")\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def get_analog_out(self, pin_number):
        """Returns specified analog output pin value.

        Args:
            pin_number (int): pin number (0-63)

        Returns:
            float: pin value
        """
        command = "get_analog_out(" + str(pin_number) + ")\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def set_analog_out(self, pin_number, pin_value):
        """Sets specified analog output pin to given value.

        Args:
            pin_number (int): pin number (0~63).
            pin_value (float): pin value
        """
        command = "set_analog_out(" + str(pin_number) + "," + str(pin_value) + ")\n"
        self.send_command(command)

    def get_joint_current(self, joint_number):
        command = "get_joint_current(" + str(joint_number) + ")\n"
        print(command)
        res = self.send_command(command)
        return self.string_to_double(res)

    def send_feed_override(self, override):
        command = "set_feed_rate(" + str(override) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_acceleration(self):
        command = "get_acceleration()\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def set_acceleration(self, acceleration):
        command = "set_acceleration(" + str(acceleration) + ")\n"
        self.send_command(command)

    def command_wait_done(self):
        command = "wait_command_done()\n"
        self.send_command(command)

    def wait(self, seconds):
        command = "wait(" + str(seconds) + ")\n"
        self.send_command(command)

    def assign_variable(self, var_name, var_value):
        command = 'assign_variable("' + str(var_name) + '",' + str(var_value) + ")\n"
        self.send_command(command)

    def get_variable(self, var_name):
        command = 'get_variable("' + str(var_name) + '")\n'
        return self.send_command(command)

    def jog_relative(self, joint_id, angle, speed, mode):
        """Relative jog.

        Example:
            jog_relative('J1', 5, 600, 1)

        Args:
            joint_id (str): 'J1' - 'J6', 'X', 'Y', 'Z', 'RX', 'RY', 'RZ'
            angle (float): relative angle or coord value to move, can be negative
                           to move to other direction
            speed (int): speed value
            mode (int): 0 (coord mode) or 1 (joint mode)

        Returns:
            str: return data from socket
        """
        command = "jog_increment({},{},{},{})\n".format(joint_id, angle, speed, mode)
        return self.send_command(command)

    def set_gripper_state(self, state, speed):
        """Sets gripper state.

        Args:
            state (int): gripper state, 0 - open, 1 - close
            speed (int): speed, 1-100

        Returns:
            str: return message
        """
        command = "set_gripper_state(" + str(state) + "," + str(speed) + ")\n"
        return self.send_command(command)

    def set_gripper_value(self, value, speed):
        """Sets gripper open/close state.

        Args:
            value (int): gripper open value, 0-100
            speed (int): speed, 1-100

        Returns:
            str: return message
        """
        command = "set_gripper_value(" + str(value) + "," + str(speed) + ")\n"
        return self.send_command(command)

    def set_gripper_calibrate(self):
        """Sets gripper zero position.

        Returns:
            str: return message
        """
        command = "set_gripper_calibrate()\n"
        return self.send_command(command)

    def set_gripper_enabled(self, enabled):
        """Set gripper enabled.

        Args:
            enabled (int): 1 - enabled, 0 - disabled

        Returns:
            str: return message
        """
        command = "set_gripper_enabled(" + str(enabled) + ")\n"
        return self.send_command(command)

    def set_gripper_mode(self, mode):
        """Sets gripper mode. Default mode is IO, after setting mode 485
           cannot set mode back to IO (need gripper reboot).

        Args:
            mode (int): 0 - 485, 1 - IO

        Returns:
            str: return message
        """
        command = "set_gripper_mode(" + str(mode) + ")\n"
        return self.send_command(command)

    def init_ele_gripper(self):
        """Initializes Electric Gripper.

        Returns:
            str: return message
        """
        command = "init_ele_gripper()\n"
        return self.send_command(command)

    def set_ele_gripper_open(self):
        """Fully opens Electric Gripper.

        Returns:
            str: return message
        """
        command = "set_ele_gripper_open()\n"
        return self.send_command(command)

    def set_ele_gripper_close(self):
        """Fully closes Electric Gripper.

        Returns:
            str: return message
        """
        command = "set_ele_gripper_close()\n"
        return self.send_command(command)

    def force_gripper_set_angle(self, angle):
        """Sets angle of Force-Controlled gripper.

        Args:
            angle (int): angle, 0-100

        Returns:
            str: return message
        """
        command = "force_gripper_set_angle(" + str(angle) + ")\n"
        return self.send_command(command)

    def force_gripper_get_angle(self):
        """Returns current angle of force-controlled gripper.

        Returns:
            str: return message
        """
        command = "force_gripper_get_angle()\n"
        return self.send_command(command)

    def force_gripper_full_open(self):
        """Fully opens force-controlled gripper.

        Returns:
            str: return message
        """
        command = "force_gripper_full_open()\n"
        return self.send_command(command)

    def force_gripper_full_close(self):
        """Fully closes force-controlled gripper.

        Returns:
            str: return message
        """
        command = "force_gripper_full_close()\n"
        return self.send_command(command)

    def force_gripper_set_torque(self, torque):
        """Sets torque of force-controlled gripper.

        Args:
            torque (int): torque, 0-100 (mapped to 100-300)

        Returns:
            str: return message
        """
        command = "force_gripper_set_torque(" + str(torque) + ")\n"
        return self.send_command(command)

    def force_gripper_get_torque(self):
        """Returns current torque of force-controlled gripper.

        Returns:
            str: return message
        """
        command = "force_gripper_get_torque()\n"
        return self.send_command(command)
    
    def set_tool_reference(self, tool_reference):
        """Set tool coordinate system.

        Args:
            tool_reference (list): 

        Returns:
            _type_: _description_
        """
        rotation_matrix = np.eye(3)  # No rotation example
        translation_vector = tool_reference[:3] # Tool offset along z-axis of flange
        rotation_matrix = CvtEulerAngleToRotationMatrix(tool_reference[3:6]* np.pi/180.0)  
        return transformation_matrix_from_parameters(rotation_matrix, translation_vector)
    
    def get_tool_coords(self):
        """Set tool coordinate system.

        Returns:
            _type_: _description_
        """
        current_coords = np.array(self.get_coords())
        tool_coords = flangeToTool(current_coords, self.tool_matrix)
        return tool_coords

    def write_tool_coords(self, tool_coords, speed):
        """Tool coordinate motion control

        Args:
            tool_coords (list): _description_
            speed (int): _description_
        """
        flange_coords = toolToflange(tool_coords, self.tool_matrix)
        self.write_coords(flange_coords, speed)


if __name__ == "__main__":
    ep = ElephantRobot("192.168.124.28", 5001)
    resp = ep.start_client()
    if resp != "":
        print(resp)
        sys.exit(1)
    print(ep.wait(5))
    print(ep.get_angles())
    print(ep.get_coords())
    print(ep.get_speed())
    # print(ep._power_on())
    # print(ep._power_off())
    print(ep.check_running())
    print(ep.state_check())
    print(ep.program_open("a.tax"))
    print(ep.program_run(0))
    print(ep.read_next_error())
    print(ep.write_coords([1, 2, 3, 4, 5, 6], 110))
    print(ep.write_coord(1, 100, 200))
    print(ep.write_angles([10, 20, 30, 40, 50, 60], 110))
    print(ep.write_angle(3, 180, 200))
    print(ep.set_speed(377))
    print(ep.set_carte_torque_limit("x", 55))
    print(ep.set_upside_down(False))
    print(ep.set_payload(100))
    print(ep._state_on())
    print(ep._state_off())
    print(ep.task_stop())
    print(ep.jog_angle("j2", 1, 300))
    print(ep.jog_coord("rY", 0, 200))
    print(ep.get_digital_in(3))
    print(ep.get_digital_out(3))
    print(ep.set_digital_out(3, 1))
    print(ep.get_digital_out(3))
    print(ep.set_analog_out(1, 3.5))
    print(ep.get_acceleration())
    print(ep.set_acceleration(55))
    print(ep.command_wait_done())
    print(ep.get_variable("f"))
    print(ep.assign_variable("ss", '"eee"'))
    print(ep.get_joint_current(1))

    print(ep.set_gripper_mode(1))
    print(ep.wait(2))
    print(ep.set_gripper_mode(0))

    print(ep.set_gripper_value(100, 20))
    print(ep.wait(2))
    print(ep.set_gripper_value(50, 20))
    print(ep.wait(2))
    print(ep.set_gripper_value(0, 20))
    print(ep.wait(2))

    print(ep.set_gripper_enabled(1))
    print(ep.wait(2))
    print(ep.set_gripper_enabled(0))

    ep.stop_client()

    name = input("input:")
    print(name)



# coding=utf-8

import locale
import numpy as np

from pymycobot.error import restrict_serial_port
from pymycobot.common import ProtocolCode, FingerGripper
from pymycobot.robot_info import _interpret_status_code
from pymycobot.close_loop import CloseLoop


class MercuryCommandGenerator(CloseLoop):
    def __init__(self, debug=False):
        super(MercuryCommandGenerator, self).__init__(debug)
        # 同步模式
        self.language, _ = locale.getdefaultlocale()
        if self.language not in ["zh_CN", "en_US"]:
            self.language = "en_US"
        self.max_joint, self.min_joint = 0, 0

    def _joint_limit_init(self):
        max_joint = np.zeros(7)
        min_joint = np.zeros(7)
        for i in range(7):
            max_joint[i] = self.get_joint_max_angle(i + 1)
            min_joint[i] = self.get_joint_min_angle(i + 1)
        return max_joint, min_joint

    def _joint_limit_judge(self, angles):
        offset = 3
        try:
            for i in range(7):
                if self.min_joint[i] + offset < angles[i] < self.max_joint[i] - offset:
                    pass
                else:
                    if self.language == "zh_CN":
                        return f"当前角度为{angles[i]}, 角度范围为： {self.min_joint[i]} ~ {self.max_joint[i]}"
                    return f"current value = {angles[i]}, limit is {self.min_joint[i]} ~ {self.max_joint[i]}"
        except TypeError:
            return "joint limit error"
        return "over limit error {}".format(angles)

    def _Singularity(self, angles):
        try:
            # Joint 6: 0 and 180 degrees are singular points
            singular_angles = [0, 180]
            state = ""
            offset = 5
            for singular in singular_angles:
                if singular - offset < angles[5] < singular + offset:
                    if self.language == "zh_CN":
                        return f"在关节 6 处检测到奇点：{angles[5]} 度"
                    return f"Singularity detected at joint 6: {angles[5]} degrees"
            return state
        except:
            return "Singularity error"

    def _check_coords(self, new_coords, is_print=0):
        try:
            first_three = new_coords[:3]
            first_three[2] -= 83.64
            info = ""
            # Calculate the Euclidean norm (magnitude)
            magnitude = np.linalg.norm(first_three)
            if is_print == 1:
                if self.language == "zh_CN":
                    info += f"当前臂展为{magnitude}, 最大的臂展为{self.arm_span}"
                else:
                    info += f"Arm span is {magnitude}, max is {self.arm_span}"

            # if magnitude > self.arm_span - 10:
            #     if self.language == "zh_CN":
            #         info += f"当前臂展为{magnitude}超出物理限位, 最大的臂展为{self.arm_span}"
            #     else:
            #         info += f"Arm span is {magnitude} exceeds physical limit, max is {self.arm_span}"
            return info
        except:
            return "check coords error"

    def _status_explain(self, status):
        error_info = _interpret_status_code(self.language, status)
        if error_info != "":
            self.arm_span = 440
        if 0x00 < status <= 0x07:
            angles = self.get_angles()
            if type(self.max_joint) == int and self.max_joint == 0:
                self.max_joint, self.min_joint = self._joint_limit_init()
            error_info += self._joint_limit_judge(angles)
        elif status in [32, 33]:
            error_coords = self.get_coords()
            error_info += self._check_coords(error_coords, 1)
        elif status == 36:
            angles = self.get_angles()
            error_info += self._Singularity(angles)

        return error_info

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        return_data = super(MercuryCommandGenerator, self)._mesg(
            genre, *args, **kwargs)
        if isinstance(return_data, tuple):
            valid_data, data_len = return_data
        elif isinstance(return_data, int):
            return return_data
        else:
            return None
        res = []
        hand_address = None
        if genre == ProtocolCode.MERCURY_GET_TOQUE_GRIPPER:
            if len(valid_data) > 1:
                hand_address = self._decode_int16(valid_data[1:3])
                valid_data = valid_data[3:]
                data_len -= 3
            else:
                return valid_data[0]
        # print(data_len, valid_data)
        if data_len in [6, 8, 12, 14, 16, 20, 24, 26, 60]:
            if data_len == 8 and (genre == ProtocolCode.IS_INIT_CALIBRATION):
                if valid_data[0] == 1:
                    return 1
                n = len(valid_data)
                for v in range(1, n):
                    res.append(valid_data[v])
            elif data_len == 8 and genre == ProtocolCode.GET_DOWN_ENCODERS:
                i = 0
                while i < data_len:
                    byte_value = int.from_bytes(
                        valid_data[i:i + 4], byteorder='big', signed=True)
                    i += 4
                    res.append(byte_value)
            elif data_len == 6 and genre in [ProtocolCode.GET_SERVO_STATUS, ProtocolCode.GET_SERVO_VOLTAGES,
                                             ProtocolCode.GET_SERVO_CURRENTS]:
                for i in range(data_len):
                    res.append(valid_data[i])
            else:
                
                for header_i in range(0, len(valid_data), 2):
                    one = valid_data[header_i: header_i + 2]
                    res.append(self._decode_int16(one))
        elif data_len == 2:
            if genre in [ProtocolCode.IS_SERVO_ENABLE]:
                return self._decode_int8(valid_data[1:2])
            elif genre in [ProtocolCode.GET_LIMIT_SWITCH]:
                for i in valid_data:
                    res.append(i)
            elif genre in [ProtocolCode.GET_ERROR_INFO]:
                res.append(self._decode_int8(valid_data[1:]))
            else:
                res.append(self._decode_int16(valid_data))
                if hand_address in [FingerGripper.GET_HAND_MAJOR_FIRMWARE_VERSION]:
                    res[0] /=10
        elif data_len == 3:
            if genre in [ProtocolCode.GET_DIGITAL_INPUTS]:
                for i in valid_data:
                    res.append(i)
            else:
                res.append(self._decode_int16(valid_data[1:]))
        elif data_len == 4:
            if genre in [ProtocolCode.COBOTX_GET_ANGLE, ProtocolCode.GET_ENCODER, ProtocolCode.GET_DYNAMIC_PARAMETERS]:
                byte_value = int.from_bytes(
                    valid_data, byteorder='big', signed=True)
                if genre in [ProtocolCode.GET_DYNAMIC_PARAMETERS]:
                    byte_value /= 1000
                res.append(byte_value)
            else:
                for i in range(1, 4):
                    res.append(valid_data[i])
        elif data_len == 7:
            # if genre ==ProtocolCode.GET_TORQUE_COMP:
            for i in range(0, data_len):
                res.append(valid_data[i])
            # else:
            #     error_list = [i for i in valid_data]
            #     for i in error_list:
            #         if i in range(16, 23):
            #             res.append(1)
            #         elif i in range(23, 29):
            #             res.append(2)
            #         elif i in range(32, 112):
            #             res.append(3)
            #         else:
            #             res.append(i)
        elif data_len in [28, 32]:  # 28 left get_zero_pos, 32 right get_zero_pos
            for i in range(0, data_len, 4):
                byte_value = int.from_bytes(
                    valid_data[i:i + 4], byteorder='big', signed=True)
                res.append(byte_value)
        elif data_len == 40 and genre == ProtocolCode.GET_ANGLES_COORDS:
            i = 0
            while i < data_len:
                if i < 28:
                    byte_value = int.from_bytes(
                        valid_data[i:i + 4], byteorder='big', signed=True)
                    res.append(byte_value)
                    i += 4
                else:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 40 and genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            # 右臂上位机错误
            i = 0
            res = []
            while i < data_len:
                if i < 10 or i >= 30:
                    res.append(valid_data[i])
                    i += 1
                elif i < 30:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 41 and genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            # 图灵右臂上位机错误
            i = 0
            res = []
            while i < data_len:
                if i < 9:
                    res.append(valid_data[i])
                    i += 1
                else:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len in [37, 46]:
            # X1 left status
            if data_len == 37:
                add_index = 9
            else:
                add_index = 10
            i = 0
            res = []
            while i < data_len:
                if i < add_index:
                    res.append(valid_data[i])
                    i += 1
                else:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 38:
            i = 0
            res = []
            while i < data_len:
                if i < 10 or i >= 30:
                    res.append(valid_data[i])
                    i += 1
                elif i < 38:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 56:
            i = 0
            while i < data_len:
                byte_value_send = int.from_bytes(
                    valid_data[i:i + 4], byteorder='big', signed=True)
                i += 4
                byte_value_current = int.from_bytes(
                    valid_data[i:i + 4], byteorder='big', signed=True)
                res.append([byte_value_send, byte_value_current])
                i += 4
        elif data_len == 48:
            i = 0
            j = 0
            res = [0, 0]
            angles = []
            coords = []
            while i < data_len:
                if i < 28:
                    byte_value_send = int.from_bytes(
                        valid_data[i:i + 4], byteorder='big', signed=True)
                    angles.append(self._int2angle(byte_value_send))
                    i += 4
                elif i < 40:

                    one = valid_data[i: i + 2]
                    one = self._decode_int16(one)
                    if j < 3:
                        one = self._int2coord(one)
                    else:
                        one = self._int2angle(one)
                    coords.append(one)
                    j += 1
                    i += 2
                else:
                    res.append(valid_data[i])
                    i += 1
            res[0] = angles
            res[1] = coords
            res[2] = 0
            res[3] = 0
        # elif genre == ProtocolCode.MERCURY_GET_TOQUE_GRIPPER:
        #     res = self._decode_int16(valid_data[-2:])
        #     address = self._decode_int16(valid_data[1:3])
        #     if address == 1:
        #         res /= 10
        else:
            if genre in [
                ProtocolCode.GET_SERVO_VOLTAGES,
                ProtocolCode.GET_SERVO_STATUS,
                ProtocolCode.GET_SERVO_TEMPS,
            ]:
                for i in range(data_len):
                    data1 = self._decode_int8(valid_data[i: i + 1])
                    res.append(0xFF & data1 if data1 < 0 else data1)
            res.append(self._decode_int8(valid_data))
        if res == []:
            return None
        if genre in [
            ProtocolCode.ROBOT_VERSION,
            ProtocolCode.GET_ROBOT_ID,
            ProtocolCode.IS_POWER_ON,
            ProtocolCode.IS_CONTROLLER_CONNECTED,
            ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
            ProtocolCode.IS_IN_POSITION,
            ProtocolCode.IS_MOVING,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.IS_ALL_SERVO_ENABLE,
            ProtocolCode.GET_SERVO_DATA,
            ProtocolCode.GET_DIGITAL_INPUT,
            ProtocolCode.GET_GRIPPER_VALUE,
            ProtocolCode.IS_GRIPPER_MOVING,
            ProtocolCode.GET_SPEED,
            ProtocolCode.GET_ENCODER,
            ProtocolCode.GET_BASIC_INPUT,
            ProtocolCode.GET_TOF_DISTANCE,
            ProtocolCode.GET_END_TYPE,
            ProtocolCode.GET_MOVEMENT_TYPE,
            ProtocolCode.GET_REFERENCE_FRAME,
            ProtocolCode.GET_FRESH_MODE,
            ProtocolCode.GET_GRIPPER_MODE,
            ProtocolCode.SET_SSID_PWD,
            # ProtocolCode.GET_ERROR_DETECT_MODE,
            ProtocolCode.POWER_ON,
            ProtocolCode.POWER_OFF,
            ProtocolCode.RELEASE_ALL_SERVOS,
            ProtocolCode.RELEASE_SERVO,
            ProtocolCode.FOCUS_ALL_SERVOS,
            ProtocolCode.FOCUS_SERVO,
            ProtocolCode.STOP,
            ProtocolCode.SET_BREAK,
            ProtocolCode.IS_BTN_CLICKED,
            ProtocolCode.GET_CONTROL_MODE,
            ProtocolCode.GET_VR_MODE,
            ProtocolCode.GET_FILTER_LEN,
            ProtocolCode.MERCURY_ERROR_COUNTS,
            ProtocolCode.GET_MAX_ACC,
            ProtocolCode.GET_MONITOR_MODE,
            ProtocolCode.GET_COLLISION_MODE,
            ProtocolCode.GET_DYNAMIC_PARAMETERS,
            ProtocolCode.GET_ERROR_INFO
        ]:
            return self._process_single(res)
        elif genre in [ProtocolCode.GET_DRAG_FIFO]:
            return [self._int2angle(angle) for angle in res]
        elif genre in [
            ProtocolCode.GET_COORDS,
            ProtocolCode.MERCURY_GET_BASE_COORDS,
            ProtocolCode.GET_TOOL_REFERENCE,
            ProtocolCode.GET_WORLD_REFERENCE,
        ]:
            if res:
                r = []
                for idx in range(3):
                    r.append(self._int2coord(res[idx]))
                for idx in range(3, 6):
                    r.append(self._int2angle(res[idx]))
                return r
            else:
                return res
        elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
            return [self._int2coord(angle) for angle in res]
        elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION]:
            return self._int2coord(self._process_single(res))
        elif genre in [
            ProtocolCode.GET_JOINT_MAX_ANGLE,
            ProtocolCode.GET_JOINT_MIN_ANGLE,
        ]:
            return self._int2coord(res[0])
        elif genre == ProtocolCode.GET_ANGLES_COORDS:
            r = []
            for index in range(len(res)):
                if index < 7:
                    r.append(self._int2angle(res[index]))
                elif index < 10:
                    r.append(self._int2coord(res[index]))
                else:
                    r.append(self._int2angle(res[index]))
            return r
        elif genre == ProtocolCode.GO_ZERO:
            r = []
            if res:
                if 1 not in res[1:]:
                    return res[0]
                else:
                    for i in range(1, len(res)):
                        if res[i] == 1:
                            r.append(i)
            return r
        elif genre in [ProtocolCode.COBOTX_GET_SOLUTION_ANGLES, ProtocolCode.MERCURY_GET_POS_OVER_SHOOT,
                       ProtocolCode.GET_SERVO_CW]:
            return self._int2angle(res[0])
        elif genre == ProtocolCode.GET_ANGLES:
            return [self._int3angle(angle) for angle in res]
        elif genre == ProtocolCode.SOLVE_INV_KINEMATICS:
            return [self._int2angle(angle) for angle in res]
        elif genre == ProtocolCode.COBOTX_GET_ANGLE:
            return self._int2angle(res[0])
        elif genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            if len(res) == 23:
                index = 9
            else:
                index = 10
            for i in range(index, len(res)):
                if res[i] != 0:
                    data = bin(res[i])[2:]
                    res[i] = []
                    while len(data) != 16:
                        data = "0" + data
                    for j in range(16):
                        if data[j] != "0":
                            error_id = 15 - j
                            res[i].append(error_id)
                    if res[i] == []:
                        res[i] = 0
            return res
        elif genre == ProtocolCode.GET_MOTORS_RUN_ERR:
            for i in range(len(res)):
                if res[i] != 0:
                    data = bin(res[i])[2:]
                    res[i] = []
                    while len(data) != 16:
                        data = "0" + data
                    for j in range(16):
                        if data[j] != "0":
                            error_id = 15 - j
                            res[i].append(error_id)
                    if res[i] == []:
                        res[i] = 0
            return res
        else:
            if isinstance(res, list) and len(res) == 1:
                return res[0]
            return res

    @restrict_serial_port
    def get_base_coords(self):
        """get base coords"""
        return self._mesg(ProtocolCode.MERCURY_GET_BASE_COORDS)

    @restrict_serial_port
    def send_base_coord(self, coord_id, base_coord, speed, _async=False):
        """Single coordinate control with the torso base as the coordinate system

        Args:
            coord_id (int): 1 to 6 correspond to x, y, z, rx, ry, rz
            base_coord (float): coord value.
                The coord range of `X` is -351.11 ~ 566.92.
                The coord range of `Y` is -645.91 ~ 272.12.
                The coord range of `Y` is -262.91 ~ 655.13.
                The coord range of `RX` is -180 ~ 180.
                The coord range of `RY` is -180 ~ 180.
                The coord range of `RZ` is -180 ~ 180.
            speed (int): 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=coord_id, base_coord=base_coord, speed=speed, serial_port=self._serial_port.port)
        if coord_id < 4:
            coord = self._coord2int(base_coord)
        else:
            coord = self._angle2int(base_coord)
        return self._mesg(ProtocolCode.MERCURY_SET_BASE_COORD, coord_id, [coord], speed, _async=_async, has_reply=True)

    @restrict_serial_port
    def send_base_coords(self, base_coords, speed, _async=False):
        """Full coordinate control

        Args:
            base_coords (list): coordinate value, [x, y, z, rx, ry, rz]
            speed (int): 1 ~ 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, base_coords=base_coords, speed=speed, serial_port=self._serial_port.port)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(base_coords[idx]))
        for angle in base_coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.MERCURY_SET_BASE_COORDS, coord_list, speed, _async=_async, has_reply=True)

    @restrict_serial_port
    def jog_base_coord(self, axis, direction, speed, _async=True):
        """Single-coordinate unidirectional motion control

        Args:
            axis (int): 1 ~ 6 correspond to [x, y, z, rx, ry, rz]
            direction (int): Direction of movement. 0 or 1.
            speed (int): 1 ~ 100.
            sync (bool): Whether to wait for the movement to complete. Default to True.

        Return:
            0: End of exercise.
            1 ~ 7: Corresponding joint exceeds limit.
            16 ~ 19: Collision protection.
            32 ~ 35: Coordinates have no adjacent solutions.
            49: Not powered on.
            50: Motor abnormality.
            51: Motor encoder error
            52: Not reaching the designated location or not reaching the designated location for more than 5 minutes (only J11, J12 available)
        """
        return self._mesg(ProtocolCode.MERCURY_JOG_BASE_COORD, axis, direction, speed, _async=_async, has_reply=True)

    @restrict_serial_port
    def set_servo_cw(self, head_id, err_angle):
        """Set the joint in-place feedback error angle

        Args:
            head_id (int): Joint ID, 11 or 12.
            err_angle (float): Error range is 0 ~ 5.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, head_id=head_id, err_angle=err_angle)
        return self._mesg(ProtocolCode.SET_SERVO_CW, head_id, [self._angle2int(err_angle)])

    @restrict_serial_port
    def get_servo_cw(self, head_id):
        """Get the joint in-place feedback error angle

        Args:
            head_id (int): Joint ID, 11 or 12.

        Returns:
            float: Error angle, range is 0 ~ 5.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, head_id=head_id)
        return self._mesg(ProtocolCode.GET_SERVO_CW, head_id)

    @restrict_serial_port
    def clear_waist_queue(self):
        """Clear the cache points of the three motors in the torso
        """
        return self._mesg(ProtocolCode.CLEAR_WAIST_QUEUE)

    @restrict_serial_port
    def jog_base_increment_coord(self, axis_id, increment, speed, _async=True):
        """Single coordinate incremental motion control

        Args:
            axis_id (int): axis id, range 1 ~ 6 corresponds to [x,y,z,rx,ry,rz]
            increment (float): Incremental value
            speed (int): speed

        Return:
            0: End of exercise.
            1 ~ 7: Corresponding joint exceeds limit.
            16 ~ 19: Collision protection.
            32 ~ 35: Coordinates have no adjacent solutions.
            49: Not powered on.
            50: Motor abnormality.
            51: Motor encoder error
            52: Not reaching the designated location or not reaching the designated location for more than 5 minutes (only J11, J12 available)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=axis_id, speed=speed)
        coord_list = []
        if axis_id < 4:
            coord_list.append(self._coord2int(increment))
        else:
            coord_list.append(self._angle2int(increment))
        return self._mesg(ProtocolCode.JOG_BASE_INCREMENT_COORD, axis_id, coord_list, speed, has_reply=True,
                          _async=_async)
        
    def is_in_position(self, data, mode=0):
        """Judge whether in the position.

        Args:
            data: A data list, angles or coords. angles len 6, coords len 6.
            mode: 0 - angles, 1 - coords, 2 - base coords

        Return:
            1 - True\n
            0 - False\n
            -1 - Error
        """
        if mode in [1,2]:
            if mode == 2:
                self.calibration_parameters(class_name=self.__class__.__name__, base_coords=data, serial_port=self._serial_port.port)
            else:
                self.calibration_parameters(class_name=self.__class__.__name__, coords=data)
            data_list = []
            for idx in range(3):
                data_list.append(self._coord2int(data[idx]))
            for idx in range(3, 6):
                data_list.append(self._angle2int(data[idx]))
        elif mode == 0:
            self.calibration_parameters(class_name=self.__class__.__name__, angles=data)
            data_list = [self._angle2int(i) for i in data]
        else:
            raise Exception("mode is not right, please input 0 or 1 or 2")
        return self._mesg(ProtocolCode.IS_IN_POSITION, data_list, mode)
    
    def write_waist_sync(self, current_angle, target_angle, speed):
        """_summary_

        Args:
            current_angle (_type_): _description_
            target_angle (_type_): _description_
            speed (_type_): _description_
        """
        self.calibration_parameters(class_name=self.__class__.__name__,
                                    current_angle=current_angle, target_angle=target_angle, speed=speed)
        return self._mesg(ProtocolCode.WRITE_WAIST_SYNC, [self._angle2int(current_angle)], [self._angle2int(target_angle)], speed)

    @restrict_serial_port
    def jog_base_rpy(self, axis, direction, speed, _async=True):
        """Rotate the end point around the fixed axis of the base coordinate system

        Args:
            axis (int): 1 ~ 3. 1 - Roll, 2 - Pitch, 3 - Yaw
            direction (int): 1 - Forward. 0 - Reverse.
            speed (int): 1 ~ 100.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, axis=axis, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_BASE_RPY, axis, direction, speed, _async=_async, has_reply=True)




# coding=utf-8

from pymycobot.common import ProtocolCode, ProGripper, FingerGripper
    
class ForceGripper:
    def calibration_parameters(self, *args, **kwargs):
        pass
    def _mesg(self, *args, **kwargs):
        pass
    # 设置力矩手爪
    def set_pro_gripper(self, gripper_id, gripper_address, value=0, has_return=False):
        # 调用校准参数函数
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id,
                                    gripper_address=gripper_address)
        # 发送设置力矩手爪的指令
        return self._mesg(ProtocolCode.MERCURY_SET_TOQUE_GRIPPER, gripper_id, [gripper_address], [value],
                          has_return=has_return)

    # 获取力矩手爪
    def get_pro_gripper(self, gripper_id, gripper_address):
        # 调用校准参数函数
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id,
                                    gripper_address=gripper_address)
        # 发送获取力矩手爪的指令
        return self._mesg(ProtocolCode.MERCURY_GET_TOQUE_GRIPPER, gripper_id, [gripper_address])

    # 设置力矩手爪角度
    def set_pro_gripper_angle(self, gripper_id, gripper_angle):
        # 调用校准参数函数
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id,
                                    gripper_angle=gripper_angle)
        # 发送设置力矩手爪角度的指令
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ANGLE, gripper_angle)

    # 设置力矩手爪打开
    def set_pro_gripper_open(self, gripper_id):
        # 调用校准参数函数
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ANGLE, 100)

    def set_pro_gripper_close(self, gripper_id):
    # 设置力矩手爪关闭
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        # 调用校准参数函数
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ANGLE, 0)
        # 发送设置力矩手爪关闭的指令

    def get_pro_gripper_angle(self, gripper_id):
    # 获取力矩手爪角度
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        # 调用校准参数函数
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_ANGLE)
        # 发送获取力矩手爪角度的指令

    def set_pro_gripper_calibration(self, gripper_id):
    # 设置力矩手爪校准
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        # 调用校准参数函数
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_CALIBRATION)
        # 发送设置力矩手爪校准的指令

    def get_pro_gripper_status(self, gripper_id):
    # 获取力矩手爪状态
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        # 调用校准参数函数
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_STATUS)
        # 发送获取力矩手爪状态的指令

    def set_pro_gripper_torque(self, gripper_id, torque):
    # 设置力矩手爪扭矩
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id, torque=torque)
        # 调用校准参数函数
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_TORQUE, torque)
        # 发送设置力矩手爪扭矩的指令

    def get_pro_gripper_torque(self, gripper_id):
    # 获取力矩手爪扭矩
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        # 调用校准参数函数
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_TORQUE)
        # 发送获取力矩手爪扭矩的指令

    def set_pro_gripper_speed(self, gripper_id, speed):
    # 设置力矩手爪速度
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id, speed=speed)
        # 调用校准参数函数
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_SPEED, speed)
        # 发送设置力矩手爪速度的指令

    def get_pro_gripper_speed(self, gripper_id):
    # 获取力矩手爪速度
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        # 调用校准参数函数
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_SPEED)
        # 发送获取力矩手爪速度的指令

    def set_pro_gripper_abs_angle(self, gripper_id, angle):
    # 设置力矩手爪绝对角度
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ABS_ANGLE, angle, has_return=True)
        # 发送设置力矩手爪绝对角度的指令

    def set_pro_gripper_pause(self, gripper_id):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_PAUSE)

    def set_pro_gripper_stop(self, gripper_id):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_STOP)

    def set_pro_gripper_resume(self, gripper_id):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_RESUME)
    

class ThreeHand:
    def calibration_parameters(self, *args, **kwargs):
        pass
    def _mesg(self, *args, **kwargs):
        pass
    def __set_tool_fittings_value(self, addr, *args, gripper_id=14, **kwargs):
        kwargs["has_replay"] = True
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)
        return self._mesg(ProtocolCode.MERCURY_SET_TOQUE_GRIPPER, gripper_id, [addr], *args or ([0x00],), **kwargs)

    def __get_tool_fittings_value(self, addr, *args, gripper_id=14, **kwargs):
        kwargs["has_replay"] = True
        return self._mesg(ProtocolCode.MERCURY_GET_TOQUE_GRIPPER, gripper_id, [addr], *args or ([0x00],), **kwargs)

    def get_hand_firmware_major_version(self, gripper_id=14):
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_MAJOR_FIRMWARE_VERSION, gripper_id=gripper_id
        )

    def get_hand_firmware_minor_version(self, gripper_id=14):
        return self.__get_tool_fittings_value(FingerGripper.GET_HAND_MINOR_FIRMWARE_VERSION, gripper_id=gripper_id)

    def set_hand_gripper_id(self, new_hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, new_hand_id=new_hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_ID, [new_hand_id], gripper_id=gripper_id
        )

    def get_hand_gripper_id(self, gripper_id=14):
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_ID, gripper_id=gripper_id
        )

    def set_hand_gripper_angle(self, hand_id, gripper_angle, gripper_id=14):
        """Set the angle of the single joint of the gripper

        Args:
            hand_id (int): 1 ~ 6
            gripper_angle (int): 0 ~ 100
            gripper_id (int) : 1 ~ 254
        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id, gripper_angle=gripper_angle)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_ANGLE, [hand_id], [gripper_angle], gripper_id=gripper_id
        )

    def get_hand_gripper_angle(self, hand_id, gripper_id=14):
        """Get the angle of the single joint of the gripper

        Args:
            hand_id (int): 1 ~ 6
            gripper_id (int) : 1 ~ 254

        Return:
            gripper_angle (int): 0 ~ 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_ANGLE, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_angles(self, angles, speed, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, speed=speed)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_ANGLES, [angles], [speed], gripper_id=gripper_id
        )

    def get_hand_gripper_angles(self, gripper_id=14):
        return self.__get_tool_fittings_value(FingerGripper.GET_HAND_ALL_ANGLES, gripper_id)

    def set_hand_gripper_torque(self, hand_id, torque, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id, torque=torque)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_TORQUE, [hand_id], [torque], gripper_id=gripper_id
        )

    def get_hand_gripper_torque(self, hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_TORQUE, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_calibrate(self, hand_id, gripper_id=14):
        """ Setting the gripper jaw zero position

        Args:
            hand_id (int): 1 ~ 6
            gripper_id (int): 1 ~ 254
        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_CALIBRATION, [hand_id], gripper_id=gripper_id
        )

    def get_hand_gripper_status(self, gripper_id=14):
        """ Get the clamping status of the gripper

        Args:
            gripper_id (int): 1 ~ 254

        Return:
            0 - Moving
            1 - Stopped moving, no clamping detected
            2 - Stopped moving, clamping detected
            3 - After clamping detected, the object fell
        """
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_STATUS, gripper_id=gripper_id
        )

    def set_hand_gripper_enabled(self, flag, gripper_id=14):
        """ Set the enable state of the gripper

        Args:
            gripper_id (int): 1 ~ 254
            flag (int): 0 or 1

        """
        self.calibration_parameters(class_name=self.__class__.__name__, flag=flag)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_ENABLED, [flag], gripper_id=gripper_id
        )

    def set_hand_gripper_speed(self, hand_id, speed, gripper_id=14):
        """ Set the speed of the gripper

        Args:
            hand_id (int): 1 ~ 6
            speed (int): 1 ~ 100
            gripper_id (int): 1 ~ 254

        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id, speed=speed)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_SPEED, [hand_id], [speed], gripper_id=gripper_id
        )

    def get_hand_gripper_default_speed(self, hand_id, gripper_id=14):
        """ Get the default speed of the gripper

        Args:
            hand_id (int): 1 ~ 6
            gripper_id (int): 1 ~ 254

        Return:
            default speed (int): 1 ~ 100

        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_DEFAULT_SPEED, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_pinch_action(self, pinch_mode, gripper_id=14):
        """ Set the pinching action of the gripper

        Args:
            gripper_id (int): 1 ~ 254
            pinch_mode (int):
                0 - Index finger and thumb pinch
                1 - Middle finger and thumb pinch
                2 - Three-finger grip
                3 - Two-finger grip
        """
        self.calibration_parameters(class_name=self.__class__.__name__, pinch_mode=pinch_mode)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_PINCH_ACTION, pinch_mode, gripper_id=gripper_id
        )

    def set_hand_gripper_p(self, hand_id, value, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_P, [hand_id], [value], gripper_id=gripper_id
        )

    def get_hand_gripper_p(self, hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_P, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_d(self, hand_id, value, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_D, [hand_id], [value], gripper_id=gripper_id
        )

    def get_hand_gripper_d(self, hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_D, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_i(self, hand_id, value, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_I, [hand_id], [value], gripper_id=gripper_id
        )

    def get_hand_gripper_i(self, hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_I, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_min_pressure(self, hand_id, value, gripper_id=14):
        """ Set the minimum starting force of the single joint of the gripper

        Args:
            hand_id (int): 1 ~ 6
            value (int): 0 ~ 254
            gripper_id (int): 1 ~ 254

        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_MIN_PRESSURE, [hand_id], [value], gripper_id=gripper_id
        )

    def get_hand_gripper_min_pressure(self, hand_id, gripper_id=14):
        """ Set the minimum starting force of the single joint of the gripper

        Args:
            gripper_id (int): 1 ~ 254
            hand_id (int): 1 ~ 6

        Return:
            min pressure value (int): 0 ~ 254

        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_MIN_PRESSURE, [hand_id], gripper_id=gripper_id
        )

    def set_hand_gripper_clockwise(self, hand_id, value, gripper_id=14):
        """
        state: 0 or 1, 0 - disable, 1 - enable
        """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_CLOCKWISE, [hand_id], [value], gripper_id=gripper_id
        )

    def get_hand_gripper_clockwise(self, hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_CLOCKWISE, hand_id, gripper_id=gripper_id
        )

    def set_hand_gripper_counterclockwise(self, hand_id, value, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__set_tool_fittings_value(
            FingerGripper.SET_HAND_GRIPPER_COUNTERCLOCKWISE, [hand_id], [value], gripper_id=gripper_id
        )

    def get_hand_gripper_counterclockwise(self, hand_id, gripper_id=14):
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_GRIPPER_COUNTERCLOCKWISE, [hand_id], gripper_id=gripper_id
        )

    def get_hand_single_pressure_sensor(self, hand_id, gripper_id=14):
        """ Get the counterclockwise runnable error of the single joint of the gripper

       Args:
           gripper_id (int): 1 ~ 254
           hand_id (int): 1 ~ 6

       Return:
           int: 0 ~ 4096

       """
        self.calibration_parameters(class_name=self.__class__.__name__, hand_id=hand_id)
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_SINGLE_PRESSURE_SENSOR, [hand_id], gripper_id=gripper_id
        )

    def get_hand_all_pressure_sensor(self, gripper_id=14):
        """ Get the counterclockwise runnable error of the single joint of the gripper

        Args:
           gripper_id (int): 1 ~ 254

        Return:
            int: 0 ~ 4096

        """
        return self.__get_tool_fittings_value(
            FingerGripper.GET_HAND_ALL_PRESSURE_SENSOR, gripper_id=gripper_id
        )

    def set_hand_gripper_pinch_action_speed_consort(self, pinch_pose, rank_mode, gripper_id=14, idle_flag=None):
        """ Setting the gripper pinching action-speed coordination

        Args:
            pinch_pose (int): 0 ~ 4
                0: All joints return to zero
                1: Index finger and thumb pinch together
                2: Middle finger and thumb pinch together
                3: Index finger and middle finger pinch together
                4: Three fingers together
            rank_mode (int): 0 ~ 5
                The degree of closure,the higher the level, the more closed
            gripper_id (int): 1 ~ 254
            idle_flag (int): default None or 1
                Idle flag. By default, there is no such byte. When this byte is 1, the idle finger can be freely manipulated.

        """

        if idle_flag is None:
            self.calibration_parameters(
                class_name=self.__class__.__name__, pinch_pose=pinch_pose, rank_mode=rank_mode
            )
            return self.__set_tool_fittings_value(
                FingerGripper.SET_HAND_GRIPPER_PINCH_ACTION_SPEED_CONSORT, pinch_pose, rank_mode, gripper_id=gripper_id
            )
        else:
            self.calibration_parameters(
                class_name=self.__class__.__name__, pinch_pose=pinch_pose, rank_mode=rank_mode, idle_flag=idle_flag
            )
            return self.__set_tool_fittings_value(
                FingerGripper.SET_HAND_GRIPPER_PINCH_ACTION_SPEED_CONSORT, pinch_pose, rank_mode, idle_flag,
                gripper_id=gripper_id
            )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from __future__ import division

import functools
import socket
import time
import threading
import serial

from pymycobot.generate import CommandGenerator
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters


def setup_serial_port(port, baudrate, timeout=None):
    serial_api = serial.Serial()
    serial_api.port = port
    serial_api.baudrate = baudrate
    serial_api.timeout = timeout
    serial_api.rts = False
    serial_api.open()
    return serial_api


def setup_socket_connect(host, port, timeout):
    socket_api = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_api.settimeout(timeout)
    socket_api.connect((host, port))
    return socket_api


class GPIOProtocolCode:
    SETUP_GPIO_MODE = 0xAA
    SETUP_GPIO_STATE = 0xAB
    SET_GPIO_OUTPUT = 0xAC
    GET_GPIO_INPUT = 0xAD


class MyCobot280RDKX5Api(CommandGenerator):

    def __init__(self, debug=False, thread_lock=True):
        super(MyCobot280RDKX5Api, self).__init__(debug)
        self.calibration_parameters = functools.partial(calibration_parameters, class_name="MyCobot280")
        self.thread_lock = thread_lock
        self.lock = threading.Lock()

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply, _ = super(MyCobot280RDKX5Api, self)._mesg(genre, *args, **kwargs)
        if self.thread_lock:
            with self.lock:
                return self._res(real_command, has_reply, genre)
        else:
            return self._res(real_command, has_reply, genre)

    def _res(self, real_command, has_reply, genre):
        if genre == ProtocolCode.SET_SSID_PWD or genre == ProtocolCode.GET_SSID_PWD:
            self._write(self._flatten(real_command))
            data = self._read(genre)
        else:
            try_count = 0
            while try_count < 3:
                self._write(self._flatten(real_command))
                data = self._read(genre)
                if data is not None and data != b'':
                    break
                try_count += 1
            else:
                return -1
        if genre == ProtocolCode.SET_SSID_PWD:
            return 1

        res = self._process_received(data, genre)

        if res is None:
            return None
        if genre in [ProtocolCode.SET_BASIC_OUTPUT]:
            return 1
        if res is not None and isinstance(res, list) and len(res) == 1 and genre not in [
            ProtocolCode.GET_BASIC_VERSION, ProtocolCode.GET_JOINT_MIN_ANGLE, ProtocolCode.GET_JOINT_MAX_ANGLE,
            ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION
        ]:
            return res[0]

        if genre in [
            ProtocolCode.IS_POWER_ON,
            ProtocolCode.IS_CONTROLLER_CONNECTED,
            ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
            ProtocolCode.IS_IN_POSITION,
            ProtocolCode.IS_MOVING,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.IS_ALL_SERVO_ENABLE,
            ProtocolCode.GET_SERVO_DATA,
            ProtocolCode.GET_DIGITAL_INPUT,
            ProtocolCode.GET_GRIPPER_VALUE,
            ProtocolCode.IS_GRIPPER_MOVING,
            ProtocolCode.GET_SPEED,
            ProtocolCode.GET_ENCODER,
            ProtocolCode.GET_BASIC_INPUT,
            ProtocolCode.GET_TOF_DISTANCE,
            ProtocolCode.GET_END_TYPE,
            ProtocolCode.GET_MOVEMENT_TYPE,
            ProtocolCode.GET_REFERENCE_FRAME,
            ProtocolCode.GET_FRESH_MODE,
            ProtocolCode.GET_GRIPPER_MODE,
            ProtocolCode.GET_ERROR_INFO,
            ProtocolCode.GET_COMMUNICATE_MODE,
            ProtocolCode.SET_COMMUNICATE_MODE,
            ProtocolCode.SetHTSGripperTorque,
            ProtocolCode.GetHTSGripperTorque,
            ProtocolCode.GetGripperProtectCurrent,
            ProtocolCode.InitGripper,
            ProtocolCode.SET_FOUR_PIECES_ZERO
        ]:
            return self._process_single(res)
        elif genre in [ProtocolCode.GET_ANGLES, ProtocolCode.SOLVE_INV_KINEMATICS]:
            return [self._int2angle(angle) for angle in res]
        elif genre in [ProtocolCode.GET_COORDS, ProtocolCode.GET_TOOL_REFERENCE, ProtocolCode.GET_WORLD_REFERENCE]:
            if res:
                r = []
                for idx in range(3):
                    r.append(self._int2coord(res[idx]))
                for idx in range(3, 6):
                    r.append(self._int2angle(res[idx]))
                return r
            else:
                return res
        elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
            return [self._int2coord(angle) for angle in res]
        elif genre in [ProtocolCode.GET_JOINT_MAX_ANGLE, ProtocolCode.GET_JOINT_MIN_ANGLE]:
            return self._int2coord(res[0])
        elif genre in [
            ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION
        ]:
            return self._int2coord(self._process_single(res))
        elif genre in [ProtocolCode.GET_REBOOT_COUNT]:
            return self._process_high_low_bytes(res)
        elif genre in (ProtocolCode.GET_ANGLES_COORDS, ):
            r = []
            for index, el in enumerate(res):
                if index < 6:
                    r.append(self._int2angle(el))
                elif index < 9:
                    r.append(self._int2coord(el))
                elif index < 12:
                    r.append(self._int2angle(el))
                else:
                    r.append(el)
            return r
        else:
            return res

    def wait(self, t):
        time.sleep(t)
        return self


class MyCobot280RDKX5CommandGenerator(MyCobot280RDKX5Api):

    # System Status
    def get_modify_version(self):
        """get modify version"""
        return self._mesg(ProtocolCode.ROBOT_VERSION)

    def get_system_version(self):
        """get system version"""
        return self._mesg(ProtocolCode.SOFTWARE_VERSION)

    def clear_queue(self):
        """Clear the command queue"""
        return self._mesg(ProtocolCode.CLEAR_COMMAND_QUEUE)

    def get_queue_length(self):
        return self._mesg(ProtocolCode.GET_COMMAND_QUEUE)

    def async_or_sync(self):
        """Set the command execution mode
        Return:
            0: Asynchronous mode
            1: Synchronous mode
        """
        return self._mesg(ProtocolCode.CHECK_ASYNC_OR_SYNC)

    def get_error_information(self):
        """Obtaining robot error information

        Return:
            0: No error message.
            1 ~ 6: The corresponding joint exceeds the limit position.
            16 ~ 19: Collision protection.
            32: Kinematics inverse solution has no solution.
            33 ~ 34: Linear motion has no adjacent solution.
        """
        return self._mesg(ProtocolCode.GET_ERROR_INFO, has_reply=True)

    def clear_error_information(self):
        """Clear robot error message"""
        return self._mesg(ProtocolCode.CLEAR_ERROR_INFO, has_reply=True)

    def set_monitor_state(self, state):
        """
        Set the monitoring state
        Args:
            state: 0 - Disable monitoring
                   1 - Enable monitoring
        """
        return self._mesg(ProtocolCode.SET_MONITOR_STATE, state)

    def get_monitor_state(self):
        """
        Get the monitoring state
        """
        return self._mesg(ProtocolCode.GET_MONITOR_STATE, has_reply=True)

    def get_reboot_count(self):
        """Get the number of times the robot has been restarted"""
        return self._mesg(ProtocolCode.GET_REBOOT_COUNT, has_reply=True)

    def power_on(self):
        """Open communication with Atom."""
        return self._mesg(ProtocolCode.POWER_ON)

    def power_off(self):
        """Close communication with Atom."""
        return self._mesg(ProtocolCode.POWER_OFF)

    def is_power_on(self):
        """Adjust robot arm status

        Return:
            1 - power on
            0 - power off
            -1 - error data
        """
        return self._mesg(ProtocolCode.IS_POWER_ON, has_reply=True)

    def release_all_servos(self, data=None):
        """Relax all joints"""
        return self._mesg(ProtocolCode.RELEASE_ALL_SERVOS)

    def focus_all_servos(self):
        """Damping all joints"""
        return self._mesg(ProtocolCode.FOCUS_ALL_SERVOS)

    def read_next_error(self):
        """Robot Error Detection

        Return:
            list len 6.
            0 : No abnormality
            1 : Communication disconnected
            2 : Unstable communication
            3 : Servo abnormality
        """
        return self._mesg(ProtocolCode.READ_NEXT_ERROR, has_reply=True)

    def set_fresh_mode(self, mode):
        """Set command refresh mode

        Args:
            mode: int.
                1 - Always execute the latest command first.
                0 - Execute instructions sequentially in the form of a queue.
        """
        self.calibration_parameters(mode=mode)
        return self._mesg(ProtocolCode.SET_FRESH_MODE, mode)

    def get_fresh_mode(self):
        """Query sports mode"""
        return self._mesg(ProtocolCode.GET_FRESH_MODE, has_reply=True)

    def set_vision_mode(self, flag):
        """Set the visual tracking mode to limit the posture flipping of send_coords in refresh mode.
        (Only for visual tracking function)

        Args:
            flag: 0/1; 0 - close; 1 - open
        """
        self.calibration_parameters(flag=flag)
        return self._mesg(ProtocolCode.SET_VISION_MODE, flag)

    def sync_send_angles(self, degrees, speed, timeout=15):
        """Send the angle in synchronous state and return when the target point is reached

        Args:
            degrees: a list of degree values(List[float]), length 6.
            speed: (int) 0 ~ 100
            timeout: default 15s.
        """
        t = time.time()
        self.send_angles(degrees, speed)
        while time.time() - t < timeout:
            f = self.is_in_position(degrees, 0)
            if f == 1:
                break
            time.sleep(0.1)
        return 1

    def sync_send_coords(self, coords, speed, mode=0, timeout=15):
        """Send the coord in synchronous state and return when the target point is reached

        Args:
            coords: a list of coord values(List[float])
            speed: (int) 0 ~ 100
            mode: (int): 0 - angular（default）, 1 - linear
            timeout: default 15s.
        """
        t = time.time()
        self.send_coords(coords, speed, mode)
        while time.time() - t < timeout:
            if self.is_in_position(coords, 1) == 1:
                break
            time.sleep(0.1)
        return 1

    def get_angles_coords(self):
        """Get joint angles and coordinates"""
        return self._mesg(ProtocolCode.GET_ANGLES_COORDS, has_reply=True)

    # JOG mode and operation
    def write_angles_time_control(self, angles, step_time):
        """Write the angle of a joint in time control mode
        Args:
            angles: Angle value
            step_time: Time unit: 30ms, range(1 ~ 255)
        """
        if step_time not in range(1, 256):
            raise ValueError("step_time must be in range(1 ~ 255)")
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.WRITE_ANGLE_TIME, angles, step_time)

    def jog_rpy(self, end_direction, direction, speed):
        """Rotate the end around a fixed axis in the base coordinate system

        Args:
            end_direction (int):  Roll, Pitch, Yaw (1-3)
            direction (int): 1 - forward rotation, 0 - reverse rotation
            speed (int): 1 ~ 100
        """
        self.calibration_parameters(end_direction=end_direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_ABSOLUTE, end_direction, direction, speed)

    def jog_increment_angle(self, joint_id, increment, speed):
        """ angle step mode

        Args:
            joint_id: int 1-6.
            increment: Angle increment value
            speed: int (0 - 100)
        """
        self.calibration_parameters(id=joint_id, speed=speed)
        return self._mesg(ProtocolCode.JOG_INCREMENT, joint_id, [self._angle2int(increment)], speed)

    def jog_increment_coord(self, axis, increment, speed):
        """coord step mode

        Args:
            axis: axis id 1 - 6.
            increment: Coord increment value
            speed: int (1 - 100)
        """
        self.calibration_parameters(id=axis, speed=speed)
        value = self._coord2int(increment) if axis <= 3 else self._angle2int(increment)
        return self._mesg(ProtocolCode.JOG_INCREMENT_COORD, axis, [value], speed)

    def set_HTS_gripper_torque(self, torque):
        """Set new adaptive gripper torque

        Args:
            torque (int): 150 ~ 980

        Return:
            0: Set failed
            1: Set successful
        """
        self.calibration_parameters(torque=torque)
        return self._mesg(ProtocolCode.SetHTSGripperTorque, [torque], has_reply=True)

    def get_HTS_gripper_torque(self):
        """Get gripper torque

        Returns:
            int: 150 ~ 980
        """
        return self._mesg(ProtocolCode.GetHTSGripperTorque, has_reply=True)

    def get_gripper_protect_current(self):
        """Get the gripper protection current

        Returns:
            int: 1 ~ 500
        """
        return self._mesg(ProtocolCode.GetGripperProtectCurrent, has_reply=True)

    def init_gripper(self):
        """Initialize gripper

        Returns:
            int: 0 or 1 (1 - success)
        """
        return self._mesg(ProtocolCode.InitGripper, has_reply=True)

    def set_gripper_protect_current(self, current):
        """Set the gripper protection current

        Args:
            current (int): 1 ~ 500
        """
        self.calibration_parameters(current=current)

        return self._mesg(ProtocolCode.SetGripperProtectCurrent, [current])

    # Atom IO
    def set_pin_mode(self, pin_no, pin_mode):
        """Set the state mode of the specified pin in atom.

        Args:
            pin_no   (int): pin.
            pin_mode (int): 0 - input, 1 - output, 2 - input pull up
        """
        self.calibration_parameters(pin_mode=pin_mode)
        return self._mesg(ProtocolCode.SET_PIN_MODE, pin_no, pin_mode)

    def get_gripper_value(self, gripper_type=None):
        """Get the value of gripper.

        Args:
            gripper_type (int): default 1
                1: Adaptive gripper
                3: Parallel gripper
                4: Flexible gripper

        Return:
            gripper value (int)
        """
        if gripper_type is None:
            return self._mesg(ProtocolCode.GET_GRIPPER_VALUE, has_reply=True)
        else:
            self.calibration_parameters(gripper_type=gripper_type)
            return self._mesg(ProtocolCode.GET_GRIPPER_VALUE, gripper_type, has_reply=True)

    def is_gripper_moving(self):
        """Judge whether the gripper is moving or not

        Returns:
            0 - not moving
            1 - is moving
            -1- error data
        """
        return self._mesg(ProtocolCode.IS_GRIPPER_MOVING, has_reply=True)

    def get_tool_system_version(self):
        """
        Read the terminal primary and minor version numbers
        """
        return self._mesg(ProtocolCode.GET_ATOM_VERSION, has_reply=True)

    def get_tool_modify_version(self):
        """
        Read the terminal modified version number
        """
        return self._mesg(ProtocolCode.OVER_LIMIT_RETURN_ZERO, has_reply=True)

    def is_tool_connected(self):
        """Check the end connection status"""
        return self._mesg(ProtocolCode.IS_CONTROLLER_CONNECTED, has_reply=True)

    def is_tool_button_click(self):
        """Check whether the button on the end is pressed"""
        return self._mesg(ProtocolCode.IS_CONTROLLER_BUTTON_PRESSED, has_reply=True)

    def set_pwm_output(self, channel, frequency, pin_val):
        """ PWM control

        Args:
            channel (int): IO number.
            frequency (int): clock frequency
            pin_val (int): Duty cycle 0 ~ 256; 128 means 50%
        """
        return self._mesg(ProtocolCode.SET_PWM_OUTPUT, channel, [frequency], pin_val)

    # communication mode
    def set_transponder_mode(self, mode):
        """Set basic communication mode

        Args:
            mode: 0 - Turn off transparent transmission，1 - Open transparent transmission
        """
        self.calibration_parameters(mode=mode)
        return self._mesg(ProtocolCode.SET_COMMUNICATE_MODE, mode, has_reply=True)

    def get_transponder_mode(self):
        return self._mesg(ProtocolCode.GET_COMMUNICATE_MODE, has_reply=True)

    # g9 servo
    def move_round(self):
        """Drive the 9g steering gear clockwise for one revolution
        """
        return self._mesg(ProtocolCode.move_round)

    def set_four_pieces_zero(self):
        """Set the zero position of the four-piece motor

        Returns:
            int: 0 or 1 (1 - success)
        """
        return self._mesg(ProtocolCode.SET_FOUR_PIECES_ZERO, has_reply=True)

    def set_servo_data(self, servo_id, data_id, value, mode=None):
        if data_id in (0, 1, 2, 3, 4):
            raise ValueError("Invalid data_id, addr 0 ~ 4 cannot be modified")
        return super().set_servo_data(servo_id, data_id, value, mode)

    # Running Status and Settings
    def set_joint_max(self, id, angle):
        """Set the joint maximum angle

        Args:
            id: int.
                Joint id 1 - 6
                for gripper: Joint id 7
            angle: 0 ~ 180
        """
        self.calibration_parameters(id=id, angle=angle)
        return self._mesg(ProtocolCode.SET_JOINT_MAX, id, angle)

    def set_joint_min(self, id, angle):
        """Set the joint minimum angle

        Args:
            id: int.
                Joint id 1 - 6
                for gripper: Joint id 7
            angle: 0 ~ 180
        """
        self.calibration_parameters(id=id, angle=angle)
        return self._mesg(ProtocolCode.SET_JOINT_MIN, id, angle)

    # servo state value
    def get_servo_speeds(self):
        """Get joint speed

        Return:
            A list unit step/s
        """
        return self._mesg(ProtocolCode.GET_SERVO_SPEED, has_reply=True)

    def get_servo_currents(self):
        """Get all joint current

        Return:
             A list unit mA
        """
        return self._mesg(ProtocolCode.GET_SERVO_CURRENTS, has_reply=True)

    def get_servo_voltages(self):
        """Get joint voltages

        Return:
            A list volts < 24 V
        """
        return self._mesg(ProtocolCode.GET_SERVO_VOLTAGES, has_reply=True)

    def get_servo_status(self):
        """Get joint status

        Return:
            [voltage, sensor, temperature, current, angle, overload], a value of 0 means no error, a value of 1 indicates an error
        """
        return self._mesg(ProtocolCode.GET_SERVO_STATUS, has_reply=True)

    def get_servo_temps(self):
        """Get joint temperature

        Return:
            A list unit ℃
        """
        return self._mesg(ProtocolCode.GET_SERVO_TEMPS, has_reply=True)

    # Coordinate transformation
    def set_tool_reference(self, coords):
        """Set tool coordinate system

        Args:
            coords: a list of coords value(List[float])
                    [x(mm), y, z, rx(angle), ry, rz]
        """
        self.calibration_parameters(coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_TOOL_REFERENCE, coord_list)

    def get_tool_reference(self):
        """Get tool coordinate system """
        return self._mesg(ProtocolCode.GET_TOOL_REFERENCE, has_reply=True)

    def set_world_reference(self, coords):
        """Set the world coordinate system

        Args:
            coords: a list of coords value(List[float])
                    [x(mm), y, z, rx(angle), ry, rz]\n
        """
        self.calibration_parameters(coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_WORLD_REFERENCE, coord_list)

    def get_world_reference(self):
        """Get the world coordinate system"""
        return self._mesg(ProtocolCode.GET_WORLD_REFERENCE, has_reply=True)

    def set_reference_frame(self, rftype):
        """Set the base coordinate system

        Args:
            rftype: 0 - base 1 - tool.
        """
        self.calibration_parameters(rftype=rftype)
        return self._mesg(ProtocolCode.SET_REFERENCE_FRAME, rftype)

    def get_reference_frame(self):
        """Get the base coordinate system

        Return:
            0 - base 1 - tool.
        """
        return self._mesg(ProtocolCode.GET_REFERENCE_FRAME, has_reply=True)

    def set_movement_type(self, move_type):
        """Set movement type

        Args:
            move_type: 1 - movel, 0 - moveJ
        """
        self.calibration_parameters(move_type=move_type)
        return self._mesg(ProtocolCode.SET_MOVEMENT_TYPE, move_type)

    def get_movement_type(self):
        """Get movement type

        Return:
            1 - movel, 0 - moveJ
        """
        return self._mesg(ProtocolCode.GET_MOVEMENT_TYPE, has_reply=True)

    def set_end_type(self, end):
        """Set end coordinate system

        Args:
            end: int
                0 - flange, 1 - tool
        """
        self.calibration_parameters(end=end)
        return self._mesg(ProtocolCode.SET_END_TYPE, end)

    def get_end_type(self):
        """Get end coordinate system

        Return:
            0 - flange, 1 - tool
        """
        return self._mesg(ProtocolCode.GET_END_TYPE, has_reply=True)

    def angles_to_coords(self, angles):
        """ Convert angles to coordinates

        Args:
            angles : A float list of all angle.

        Return:
            list: A float list of all coordinates.
        """
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.GET_COORDS, angles, has_reply=True)

    def solve_inv_kinematics(self, target_coords, current_angles):
        """ Convert target coordinates to angles

        Args:
            target_coords: A float list of all coordinates.
            current_angles : A float list of all angle.

        Return:
            list: A float list of all angle.
        """
        angles = [self._angle2int(angle) for angle in current_angles]
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(target_coords[idx]))
        for angle in target_coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SOLVE_INV_KINEMATICS, coord_list, angles, has_reply=True)

    def is_torque_gripper(self):
        """Whether it is a force-controlled gripper

        Return:
            40 - Force control
            9 - Non-force control
        """
        return self.get_servo_data(7, 1)

    def set_gripper_state(self, flag, speed, _type_1=None, is_torque=None):
        """Set gripper switch state

        Args:
            flag  (int): 0 - open, 1 - close, 254 - release
            speed (int): 1 ~ 100
            _type_1 (int): default 1
                1 : Adaptive gripper. default to adaptive gripper
                2 : 5 finger dexterous hand
                3 : Parallel gripper, this parameter can be omitted
                4 : Flexible gripper
            is_torque (int): When there is no type parameter, this parameter can be omitted.
                1: Force control
                0: Non-force control
        """
        self.calibration_parameters(flag=flag, speed=speed, _type_1=_type_1, is_torque=is_torque)
        args = [flag, speed]
        if _type_1 is not None:
            args.append(_type_1)
        if is_torque is not None:
            args.append(is_torque)
        return self._mesg(ProtocolCode.SET_GRIPPER_STATE, *args)

    def set_gripper_value(self, gripper_value, speed, gripper_type=None, is_torque=None):
        """Set gripper value

        Args:
            gripper_value (int): 0 ~ 100
            speed (int): 1 ~ 100
            gripper_type (int): default 1
                1: Adaptive gripper
                3: Parallel gripper, this parameter can be omitted
                4: Flexible gripper
            is_torque (int): When there is no type parameter, this parameter can be omitted.
                1: Force control
                0: Non-force control
        """
        self.calibration_parameters(gripper_value=gripper_value, speed=speed,
                                    gripper_type=gripper_type, is_torque=is_torque)
        args = [gripper_value, speed]
        if gripper_type is not None:
            args.append(gripper_type)
        if is_torque is not None:
            args.append(is_torque)
        return self._mesg(ProtocolCode.SET_GRIPPER_VALUE, *args, has_reply=True)

    def drag_start_record(self):
        """Start track recording

        Return:
            Recording queue length
        """

        return self._mesg(ProtocolCode.DRAG_START_RECORD, has_reply=True)

    def drag_end_record(self):
        """End track recording

        Return:
             Recording queue length
        """

        return self._mesg(ProtocolCode.DRAG_END_RECORD, has_reply=True)

    def drag_get_record_data(self):
        """Get the recorded track

        Return:
            List of potential values (encoder values) and operating speeds of each joint
            eg: [J1_encoder, J1_run_speed,J2_encoder, J2_run_speed,J3_encoder, J3_run_speed,J4_encoder, J4_run_speed,J5_
            encoder, J5_run_speed,J6_encoder, J6_run_speed]
        """

        return self._mesg(ProtocolCode.DRAG_GET_RECORD_DATA, has_reply=True)

    def drag_get_record_len(self):
        """Get the total number of recorded points

        Return:
            Recording queue length
        """

        return self._mesg(ProtocolCode.DRAG_GET_RECORD_LEN, has_reply=True)

    def drag_clear_record_data(self):
        """Clear recording track

        Return:
            Recording queue length 0
        """

        return self._mesg(ProtocolCode.DRAG_CLEAR_RECORD_DATA, has_reply=True)


class MyCobot280RDKX5(MyCobot280RDKX5CommandGenerator):

    def __init__(self, port, baudrate=100_0000, timeout=0.1, debug=False, thread_lock=True):
        """
        Args:
            port     : port string
            baudrate : baud rate int, default 100_0000
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super().__init__(debug, thread_lock)
        self._serial_port = setup_serial_port(port=port, baudrate=baudrate, timeout=timeout)
        self._write = functools.partial(write, self)
        self._read = functools.partial(read, self)

    def close(self):
        self._serial_port.close()

    def open(self):
        self._serial_port.open()


class MyCobot280RDKX5Socket(MyCobot280RDKX5CommandGenerator):
    """MyCobot 280 RDK X5 Socket Control Class

    server file: https://github.com/elephantrobotics/pymycobot/demo/Server_280_RDK_X5.py
    """
    def __init__(self, ip, port=30002, timeout=0.1, debug=False, thread_lock=True):
        super().__init__(debug, thread_lock)
        self.sock = setup_socket_connect(ip, port, timeout)
        self._write = functools.partial(write, self, method="socket")
        self._read = functools.partial(read, self, method="socket")

    def set_gpio_mode(self, mode):
        """Set pin coding method
        Args:
            mode: (int) 0 - BCM, 1 - BOARD

        returns:
            (int) 1 - success, 255 - error
        """
        if mode not in (0, 1):
            raise ValueError("mode must be 0 or 1")
        return self._mesg(GPIOProtocolCode.SETUP_GPIO_MODE, mode)

    def setup_gpio_state(self, pin_no, mode, initial=1):
        """Set the pin as input or output
        Args:
            pin_no: (int) pin id
            mode: (int) 0 - input, 1 - output
            initial: (int) 0 - low, 1 - high
        returns:
            (int) 1 - success, 255 - error
        """
        if mode not in (0, 1):
            raise ValueError("mode must be 0 or 1")

        if initial not in (0, 1):
            raise ValueError("initial must be 0 or 1")

        return self._mesg(GPIOProtocolCode.SETUP_GPIO_STATE, pin_no, mode, initial)

    def set_gpio_output(self, pin_no, state):
        """Set the pin to high or low level
        Args:
            pin_no: (int) pin id.
            state: (int) 0 - low, 1 - high
        returns:
            (int) 1 - success, 255 - error
        """
        return self._mesg(GPIOProtocolCode.SET_GPIO_OUTPUT, pin_no, state)

    def get_gpio_input(self, pin_no):
        """Get pin level status.
        Args:
            pin_no: (int) pin id.
        Returns:
            (int) 0 - low, 1 - high, 255 - error
        """
        return self._mesg(GPIOProtocolCode.GET_GPIO_INPUT, pin_no, has_reply=True)

    def close(self):
        self.sock.close()


def main():
    rdx_x5 = MyCobot280RDKX5("Com30", debug=True)
    print(rdx_x5.get_reboot_count())
    # mc_sock = MyCobot280RDKX5Socket('192.168.1.246', port=30002, debug=True)
    # print(mc_sock.send_angle(1, 100, 50))
    # print(mc_sock.get_quick_move_message())
    # print(mc_sock.set_gpio_mode(0))
    # print(mc_sock.setup_gpio_state(5, 1, initial=1))
    # print(mc_sock.set_gpio_output(5, 0))
    # print(mc_sock.get_gpio_input(5))


if __name__ == '__main__':
    main()
